/**************************************************************************/
/*                                                                        */
/*  Thibaut Balabonski, Sylvain Conchon, Jean-Christophe Filliâtre,       */
/*  Kim Nguyen, Laurent Sartre                                            */
/*                                                                        */
/*  Informatique - MP2I/MPI - CPGE 1re et 2e années.                      */
/*  Cours et exercices corrigés. Éditions Ellipses, 2022.                 */
/*                                                                        */
/*  https://www.informatique-mpi.fr/                                      */
/*                                                                        */
/**************************************************************************/

#include <stdio.h>
#include <stdbool.h>
#include <pthread.h>
#include <semaphore.h>
#include "queue.h"

#define fence() __asm__ __volatile__ ("mfence":::"memory")

queue *buffer;

sem_t empty;
sem_t full;
pthread_mutex_t m = PTHREAD_MUTEX_INITIALIZER;

int size = 5;
int nb_p = 2;
int nb_c = 3;

void *producer(void *arg){
  int i =  *((int *)arg);
  printf("Start producer %d\n", i);
  while (1) {
    sem_wait(&empty);
    pthread_mutex_lock(&m);
    int v = rand() % 100;
    printf("Producer %d : %d\n", i, v);
    queue_enqueue(buffer, v);
    pthread_mutex_unlock(&m);
    sem_post(&full);
    sched_yield();
  }
}

void *consumer(void *arg){
  int i =  *((int *)arg);
  printf("Start consummer %d\n", i);
  while (1) {
    sem_wait(&full);
    pthread_mutex_lock(&m);
    int v = queue_dequeue(buffer);
    printf("Consumer %d : %d\n", i, v);
    pthread_mutex_unlock(&m);
    sem_post(&empty);
    sched_yield();
  }
}

int main(int argc, char **argv) {
  buffer = queue_create();
  sem_init(&empty, 0, size);
  sem_init(&full,0, 0);

  pthread_t *p = malloc(nb_p * sizeof(pthread_t));
  int *pid = malloc(nb_p * sizeof(int));

  pthread_t *c = malloc(nb_c * sizeof(pthread_t));
  int *cid = malloc(nb_p * sizeof(int));

  for (int i = 0; i < nb_p; i++) {
    pid[i] = i;
    pthread_create(&p[i], NULL, producer, &pid[i]);
  }
  for (int i = 0; i < nb_c; i++) {
    cid[i] = i;
    pthread_create(&c[i], NULL, consumer, &cid[i]);
  }
  pthread_join(c[0], NULL);
  return 0;
}
